(******************************************************************************
 * Mathias Bourgoin, Universit Pierre et Marie Curie (2011)
 *
 * Mathias.Bourgoin@gmail.com
 *
 * This software is a computer program whose purpose is to allow
 * GPU programming with the OCaml language.
 *
 * This software is governed by the CeCILL-B license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info". 
 * 
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability. 
 * 
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
*******************************************************************************)
open Spoc
val kernel_max :
  ('a, 'b) Vector.vector * ('a, 'b) Vector.vector * int ->
  Kernel.block * Kernel.grid ->
  int -> Devices.device -> Kernel.kernel -> unit
class ['a, 'b] class_kernel_max :
  object
    constraint 'a = float
    constraint 'b = Bigarray.float64_elt
    val binaries : (Devices.device, Kernel.kernel) Hashtbl.t
    val mutable cuda_sources : string list
    val file_file : string
    val kernel_name : string
    val mutable opencl_sources : string list
    method args_to_list :
      ('a, 'b) Vector.vector * ('a, 'b) Vector.vector * int ->
      ('a, 'b) Kernel.kernelArgs array
    method compile : ?debug:bool -> Devices.device -> unit
    method compile_and_run :
      ('a, 'b) Vector.vector * ('a, 'b) Vector.vector * int ->
      Kernel.block * Kernel.grid ->
      ?debug:bool -> int -> Devices.device -> unit
    method exec :
      ('a, 'b) Vector.vector * ('a, 'b) Vector.vector * int ->
      Kernel.block * Kernel.grid ->
      int -> Devices.device -> Kernel.kernel -> unit
    method get_cuda_sources : unit -> string list
    method get_opencl_sources : unit -> string list
    method list_to_args :
      ('a, 'b) Kernel.kernelArgs array ->
      ('a, 'b) Vector.vector * ('a, 'b) Vector.vector * int
    method reload_sources : unit -> unit
    method run :
      ('a, 'b) Vector.vector * ('a, 'b) Vector.vector * int ->
      Kernel.block * Kernel.grid -> int -> Devices.device -> unit
    method set_cuda_sources : string -> unit
    method set_opencl_sources : string -> unit
    method reset_binaries : unit->unit
  end
val spoc_max : (float, Bigarray.float64_elt) class_kernel_max
class ['a, 'b] class_dummy_kernel :
  object
    val binaries : (Devices.device, Kernel.kernel) Hashtbl.t
    val mutable cuda_sources : string list
    val file_file : string
    val kernel_name : string
    val mutable opencl_sources : string list
    method args_to_list : 'a -> 'b
    method compile : ?debug:bool -> Devices.device -> unit
    method compile_and_run :
      'a ->
      Kernel.block * Kernel.grid ->
      ?debug:bool -> int -> Devices.device -> unit
    method exec :
      'a ->
      Kernel.block * Kernel.grid ->
      int -> Devices.device -> Kernel.kernel -> unit
    method get_cuda_sources : unit -> string list
    method get_opencl_sources : unit -> string list
    method list_to_args : 'b -> 'a
    method reload_sources : unit -> unit
    method run :
      'a -> Kernel.block * Kernel.grid -> int -> Devices.device -> unit
    method set_cuda_sources : string -> unit
    method set_opencl_sources : string -> unit
    method reset_binaries : unit->unit
  end
type reduction = Max
type application = Map | Pipe | Reduce | Par | Filter
class type ['a, 'b, 'c] app =
  object
    constraint 'c = ('d, 'e) Vector.vector
    val env : 'a
    val ker : ('a, 'b) Kernel.spoc_kernel
    val kind : application
    method env : unit -> 'a
    method ker : unit -> ('a, 'b) Kernel.spoc_kernel
    method kind : unit -> application
    method par_run : Devices.device list -> ('f, 'g) Vector.vector -> 'c
    method run :
      ?queue_id:int -> Devices.device -> ('h, 'i) Vector.vector -> 'c
  end
class ['a, 'b, 'c] map :
  ('a, 'b) #Kernel.spoc_kernel ->
  'c ->
  'a ->
  object
    constraint 'b = ('d, 'e) Kernel.kernelArgs array
    constraint 'c = ('d, 'e) Vector.vector
    val env : 'a
    val ker : ('a, 'b) Kernel.spoc_kernel
    val kind : application
    method env : unit -> 'a
    method ker : unit -> ('a, 'b) Kernel.spoc_kernel
    method kind : unit -> application
    method par_run : Devices.device list -> ('f, 'g) Vector.vector -> 'c
    method run :
      ?queue_id:int -> Devices.device -> ('h, 'i) Vector.vector -> 'c
  end
val get_vector : ('a, 'b) Kernel.kernelArgs -> ('a, 'b) Vector.vector
val transfer_if_vector :
  ('a, 'b) Kernel.kernelArgs -> Devices.device -> int -> unit
class ['a, 'b, 'c] pipe :
  ('a, 'b, ('d, 'e) Vector.vector) #app ->
  ('f, ('g, 'h) Kernel.kernelArgs array, 'c) #app ->
  object
    constraint 'b = ('k, 'l) Kernel.kernelArgs array
    constraint 'c = ('i, 'j) Vector.vector
    val env : 'a
    val ker : ('a, 'b) Kernel.spoc_kernel
    val kind : application
    method env : unit -> 'a
    method ker : unit -> ('a, 'b) Kernel.spoc_kernel
    method kind : unit -> application
    method par_run : Devices.device list -> ('m, 'n) Vector.vector -> 'c
    method run :
      ?queue_id:int -> Devices.device -> ('o, 'p) Vector.vector -> 'c
  end
class ['a, 'b, 'c] reduce :
  ('a, 'b) #Kernel.spoc_kernel ->
  'c ->
  'a ->
  object
    constraint 'b = ('f, 'g) Kernel.kernelArgs array
    constraint 'c = ('d, 'e) Vector.vector
    val env : 'a
    val ker : ('a, 'b) Kernel.spoc_kernel
    val kind : application
    method env : unit -> 'a
    method ker : unit -> ('a, 'b) Kernel.spoc_kernel
    method kind : unit -> application
    method par_run : Devices.device list -> ('h, 'i) Vector.vector -> 'c
    method run :
      ?queue_id:int -> Devices.device -> ('j, 'k) Vector.vector -> 'c
  end
val run :
  ('a, 'b, ('c, 'd) Vector.vector) #app ->
  ?queue_id:int ->
  Devices.device -> ('e, 'f) Vector.vector -> ('c, 'd) Vector.vector
val par_run :
  ('a, 'b, ('c, 'd) Vector.vector) #app ->
  Devices.device list -> ('e, 'f) Vector.vector -> ('c, 'd) Vector.vector
val pipe :
  ('a, ('b, 'c) Kernel.kernelArgs array, ('d, 'e) Vector.vector) #app ->
  ('f, ('g, 'h) Kernel.kernelArgs array, ('i, 'j) Vector.vector) #app ->
  ('a, ('b, 'c) Kernel.kernelArgs array, ('i, 'j) Vector.vector) pipe
val map :
  ('a, ('b, 'c) Kernel.kernelArgs array) #Kernel.spoc_kernel ->
  ('b, 'c) Vector.vector ->
  'a -> ('a, ('b, 'c) Kernel.kernelArgs array, ('b, 'c) Vector.vector) map
val reduce :
  ('a, ('b, 'c) Kernel.kernelArgs array) #Kernel.spoc_kernel ->
  ('d, 'e) Vector.vector ->
  'a -> ('a, ('b, 'c) Kernel.kernelArgs array, ('d, 'e) Vector.vector) reduce
