open Kirc_Ast

let space i =
  String.make i ' '

let return_v = ref ("","")

let rec parse i = function
  | Kern (args,body) -> 
    (let pargs = parse i args in
     let pbody = 
       match body with 
       | Seq (_,_)  -> (parse i body)
       | _  ->  ((parse i body)^"\n"^(space i))
     in
     (pargs ^ pbody)^ "\n}\n#ifdef __cplusplus\n}\n#endif")
  | Local (x,y)  -> (space (i))^"{"^
    (parse i x)^";\n"^
    (space (i))^(parse i y)^
    ";\n"^(space (i))^"}\n"
  | VecVar (t,i)  -> 
    (match t with
    | Int _ ->"int"
    | Float _ -> "float"
    | Double _ -> "double"
    | _ -> assert false
    )^("* spoc_var"^(string_of_int (i)))
  | IdName s  ->  s
  | IntVar s -> ("int spoc_var"^(string_of_int s))
  | FloatVar s -> ("float spoc_var"^(string_of_int s))
  | UnitVar v -> assert false
  | CastDoubleVar s -> ("(double) spoc_var"^(string_of_int s))
  | DoubleVar s -> ("double spoc_var"^(string_of_int s))
  | Params k -> 
    ("#ifdef __cplusplus\nextern \"C\" {\n#endif\n\n__global__ void spoc_dummy ( "^
	(if (fst !return_v) <> "" then
	    (fst !return_v)^", " else "")^(parse i k)^" ) {\n")
  (*    let rec aux acc = function
	| t::[]  ->  acc^", " ^(parse t)
	| t::q  -> aux (acc^","^parse t) q
	in
	"f (" ^(List.fold_left (fun a b -> (a^(parse b))) "" l) ^ ")\n{"*)
  |Concat (a,b)  ->  
    (match b with 
    | Empty  -> (parse i a)
    | Concat (c,d)  ->  ((parse i a)^", "^(parse i b))
    | _  -> failwith "parse concat"
    )
  | Plus  (a,b) -> ("("^(parse_int i a)^" + "^(parse_int i b)^")")
  | Plusf  (a,b) -> ("("^(parse_float i a)^" + "^(parse_float i b)^")")
  | Min  (a,b) -> ("("^(parse_int i a)^" - "^(parse_int i b)^")")
  | Minf  (a,b) -> ("("^(parse_float i a)^" - "^(parse_float i b)^")")
  | Mul  (a,b) -> ("("^(parse_int i a)^" * "^(parse_int i b)^")")
  | Mulf  (a,b) -> ("("^(parse_float i a)^" * "^(parse_float i b)^")")
  | Div  (a,b) -> ("("^(parse_int i a)^" / "^(parse_int i b)^")")
  | Divf  (a,b) -> ("("^(parse_float i a)^" / "^(parse_float i b)^")")
  | Mod  (a,b) -> ("("^(parse_int i a)^" % "^(parse_int i b)^")")

  | Id (s)  -> s
  | Set (var,value) 
  | Acc (var,value) -> 
    ((parse i var)^" = "^
        (parse i value)^";")
  | Decl (var) ->
    (parse i var)
  | SetLocalVar (v,gv,k) -> 
    ((parse i v)^" = "^
	(match gv with
	| Intrinsics i -> ((parse_intrinsics i)^";")
	| _ -> (parse i gv))^";\n"^(space i)^(parse i k))
  | Return k -> 
    (if (snd !return_v) <> "" then
	snd !return_v
     else
	"")^
      (parse i k)
  | Unit  -> ""
  | IntVecAcc (vec,idx)  -> (parse i vec)^"["^(parse i idx)^"]"
  | SetV (vecacc,value)  -> (
    (parse i vecacc)^" = "^(parse i value)^";") 
  | Int  a  -> string_of_int a
  | GInt  a  -> string_of_int (a ())
  | GFloat  a  -> (string_of_float (a ()))^"f"
  | Float f -> (string_of_float f)^"f"
  | Double f -> string_of_float f
  | IntId (s,_) -> s
  |Intrinsics gv -> parse_intrinsics gv
  | Seq (a,b)  -> (parse i a)^" ;\n"^(space i)^(parse i b)
  | Ife (a,b,c) -> "if ("^(parse i a)^"){\n"^(space (i+2))^(parse (i+2) b)^";}\n"^(space i)^"else{\n"^(space (i+2))^(parse (i+2) c)^";}\n"^(space i)
  | If (a,b) -> "if ("^(parse i a)^"){\n"^(space (i+2))^(parse (i+2) b)^";}\n"^(space i)
  | Or (a,b) -> (parse i a)^" || "^(parse i b)
  | And (a,b) -> (parse i a)^" && "^(parse i b)
  | EqBool (a,b) -> (parse i a)^" == "^(parse i b)
  | LtBool (a,b) -> (parse i a)^" < "^(parse i b)
  | GtBool (a,b) -> (parse i a)^" > "^(parse i b)
  | LtEBool (a,b) -> (parse i a)^" <= "^(parse i b)
  | GtEBool (a,b) -> (parse i a)^" >= "^(parse i b)
  | DoLoop (a,b,c,d) -> 
    let id = parse i a in
    let min = parse i b in
    let max = parse i c in
    let body = parse (i+2) d in
    "for (int "^id^" = "^min^"; "^id^" <= "^max^"; "^id^"++){\n"^(space (i+2))^body^"}"
  | While (a,b) ->
    let cond = parse i a in
    let body = parse (i+2) b in
    "while ("^cond^"){\n"^(space (i+2))^body^"}"
  | App (a,b) ->
    let f = parse i a in
    let rec aux = function
			| t::[] -> parse i t
      | t::q -> (parse i t)^","^(aux q)
			| [] -> assert false
    in 
		(match a with 
		| Intrinsics ("return","return") -> f^" "^(aux (Array.to_list b))^" "
		|  _ -> f^" ("^(aux (Array.to_list b))^") ")
  | Empty  -> ""

and parse_int n = function
  | IntId (s,_) -> s
  | Int i  ->  string_of_int i
  | GInt i  ->  string_of_int (i ())
  | IntVecAcc (s,i)  -> (parse n s)^"["^(parse_int n i)^"]"
  | Plus (a,b) as v ->  parse n v
  | Min (a,b) as v ->  parse n v
  | Mul (a,b) as v ->  parse n v
  | Mod (a,b) as v ->  parse n v
  | Div (a,b) as v ->  parse n v
  | App (a,b) as v -> parse n v
  | a -> parse_float n a
(*  | _  -> failwith "error parse_int" *)

and parse_float n = function
  | IntId (s,_) -> s
  | Float f  ->  (string_of_float f)^"f"
  | GFloat f  ->  (string_of_float (f ()))^"f"
  | CastDoubleVar s -> ("(double) spoc_var"^(string_of_int s))
  | Double f  ->  "(double) "^(string_of_float f)
  | IntVecAcc (s,i)  -> (parse n s)^"["^(parse_int n i)^"]"
  | Plusf (a,b) as v ->  parse n v
  | Minf (a,b) as v ->  parse n v
  | Mulf (a,b) as v ->  parse n v
  | Divf (a,b) as v ->  parse n v
  | App (a,b) as v -> parse n v
  | Intrinsics gv -> parse_intrinsics gv
  | a  -> print_ast a; failwith "cuda error parse_float"

and parse_vect = function
  | IntVect i  -> i
  | _  -> failwith "error parse_vect"

and parse_intrinsics (cudas,opencls) = cudas
